<?php

/**
 * Version numbers
 */
class Less_Version {

	/* Current release version of less.php */
	public const version = '4.2.1';

	/* Upstream less.js version that this release should be compatible with */
	public const less_version = '2.5.3';

	/* Parser cache version */
	public const cache_version = '253-2';
}
